#!/usr/bin/env pypy

###############################################################################
# Mjöllnir -- a simulator of the magnetic properties of polynuclear complexes
# Copyright (c) 2009--2020 Anton S. Lytvynenko, Elena A. Mikhalyova
#
# The authors thanks to Dr. V. A. Bogaenko and to Sergei Litvinenko for
# technical help and valuable discussion.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################

import sys, itertools
from math import sqrt
from operator import itemgetter


def fl_equal(f1,f2):
    '''A simple function that implements a correct check for (quasi)equality of floats.'''
    return abs(f1-f2)<sys.float_info.epsilon

def find_ps_differ(psi,psj,num,sign):
    '''Checks conditions on psi1 and psi2 vectors in <psi1|operator|psi2> matrix element
        construction: all elements must be equal but _num_'s should differ by 1 or -1 (sign).'''
    res = True
    for i in range(len(psi)):
        if i!=num and not fl_equal(psi[i],psj[i]):
            res = False
            break;
        elif i==num and not fl_equal(float(psi[i])-float(psj[i]),sign):
            res = False
            break;
    return res

def find_ps_2_differs(psi,psj,num1,num2,sign1,sign2):
    '''Checks conditions on psi1 and psi2 vectors in <psi1|operator|psi2> matrix element construction: 
    all elements must be equal but _num1_'s and _num2_'s should differ by 1 or -1 (sign1,sign2).'''
    res = True
    for i in range(len(psi)):
        if i!=num1 and i!=num2 and psi[i]!=psj[i]:
            res = False
            break;
        elif i==num1 and not fl_equal(float(psi[i])-float(psj[i]),sign1):
            res = False
            break;
        elif i==num2 and not fl_equal(float(psi[i])-float(psj[i]),sign2):
            res = False
            break;
    return res

def filter_zeros(matrix):
    return [(i,j,x) for i,j,x in matrix if not fl_equal(x,0.0)]

def merge_sum(matrices):
    sum_matrix = [item for matrix in matrices for item in matrix] # merge the matrices
    # itertools.chain.from_iterable() won't do -- it will destroy the inner representation of the items too.
    sum_matrix.sort(key=itemgetter(0,1))
    summands = itertools.groupby(sum_matrix,itemgetter(0,1))
    sum_values = lambda acc,i: acc + i[2]
    return filter_zeros([(key[0],key[1], reduce(sum_values,elements,0.0)) for key, elements in summands])

def merge_by_coef(coef_matrices):
    coef_matrices.sort(key=itemgetter(0))
    summands = itertools.groupby(coef_matrices,itemgetter(0))
    merged_coef_matrices = []
    for coef, coef_matrices_group in summands:
        merged_coef_matrices.append((coef,merge_sum([cm[1] for cm in coef_matrices_group])))
    return merged_coef_matrices

def z(ham_vectors, affected):
    return [(i,i,ham_vectors[i][affected]) for i in range(len(ham_vectors))]

def z_square(ham_vectors, affected):
    return [(i,i,(ham_vectors[i][affected])**2) for i in range(len(ham_vectors))]

def z1z2(ham_vectors, affected1,affected2):
    return [(i,i,ham_vectors[i][affected1]*ham_vectors[i][affected2]) for i in range(len(ham_vectors))]


def plus(ham_vectors, affected, LSJ):
    msl = lambda x: ham_vectors[x][affected]
    return [(i,j,sqrt(LSJ*(LSJ+1) - msl(j)*(msl(j)+1))) for i in range(len(ham_vectors)) 
            for j in range(len(ham_vectors)) if find_ps_differ(ham_vectors[i],ham_vectors[j],affected,1)]

def minus(ham_vectors, affected, LSJ):
    msl = lambda x: ham_vectors[x][affected]
    return [(i,j,sqrt(LSJ*(LSJ+1) - msl(j)*(msl(j)-1))) for i in range(len(ham_vectors)) 
            for j in range(len(ham_vectors)) if find_ps_differ(ham_vectors[i],ham_vectors[j],affected,-1)]

def plusminus(ham_vectors, affected1, affected2, LSJ1, LSJ2):
    msl1 = lambda x: ham_vectors[x][affected1]
    msl2 = lambda x: ham_vectors[x][affected2]
    l = len(ham_vectors)
    return [(i,j,sqrt(LSJ1*(LSJ1+1) - msl1(j)*(msl1(j)+1))*sqrt(LSJ2*(LSJ2+1) - msl2(j)*(msl2(j)-1)))
            for i in range(l) for j in range(l) if find_ps_2_differs(ham_vectors[i],ham_vectors[j],affected1,affected2,1,-1)]

def minusplus(ham_vectors, affected1, affected2, LSJ1, LSJ2):
    msl1 = lambda x: ham_vectors[x][affected1]
    msl2 = lambda x: ham_vectors[x][affected2]
    l = len(ham_vectors)
    return [(i,j,sqrt(LSJ1*(LSJ1+1) - msl1(j)*(msl1(j)-1))*sqrt(LSJ2*(LSJ2+1) - msl2(j)*(msl2(j)+1)))
            for i in range(l) for j in range(l) if find_ps_2_differs(ham_vectors[i],ham_vectors[j],affected1,affected2,-1,1)]

def const_by_matrix(coef, matrix):
    return [(i,j,coef*t) for i,j,t in matrix]
