#!/usr/bin/env pypy

###############################################################################
# Mjöllnir -- a simulator of the magnetic properties of polynuclear complexes
# Copyright (c) 2009--2020 Anton S. Lytvynenko, Elena A. Mikhalyova
#
# The authors thanks to Dr. V. A. Bogaenko and to Sergei Litvinenko for
# technical help and valuable discussion.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################

import logging

from mj_config import *

from mj_matrix import *
from mj_lanthanides import *
from mj_equations import *
from mj_userdef import *

#class mj_fragment:
#    'Abstract class : a distinct fragment of Mjollir model. Specific cases of fragment treatment MUST be described in ancestor classes'
#    # Instance variables:
#    # frag_num --- int
#    # frag_type --- string
#    pass

def parse_fragment_file(fname):
    'This routine parses the file with fragments list and initially checks it'
    logging.info("Trying to parse fragment file")
    fin = open(fname)
    fin_lines = map(lambda x: x.strip(),fin.readlines()) # One fragment per line, skipping possible whitespaces
    frags_list = []
    n = 0 # Counter to check the order of fragments' numbers
    for i in fin_lines:
        if i[:1] == "#": # Lines with "#" as the first non-whitespace character are considered comments
            pass
        else:
            frag_num_str,frag_type,frag_name,frag_comment = i.split()
            frag_num = int(frag_num_str)
            if int(frag_num) != n:
                raise Exception("parse_fragment_file (line "+str(n)+ " of the file): fragments must be listed sequentially in ascending order")
            if frag_type not in ["matrix","lanthanide","equation","userdef"]:
                raise Exception("parse_fragment_file (line "+str(n)+
                    " of the file): fragment type must be one of \"matrix\", \"lanthanide\", \"equation\" or \"userdef\"")
            n = n+1
            frags_list.append((frag_num,frag_type,frag_name,frag_comment))
    logging.info("Done with parsing the fragment file")
    return frags_list

class mj_model:
    'This class represents the entire magnetochemical model of the compound'
    # Instance variables:
    # experimental --- array, experimental data ([T, xT] pairs)
    # fragments --- arr of fragment instances [type mj_fragment])
    def __init__(self):
        # First we parse the list of fragments within the model
        # We expect and array such as: [(0, 'matrix', 'Fe2Ni', 'comment1'), (1, 'lanthanide', 'Eu', 'comment1'), 
        # (2, 'equation', 'Cu2', 'comment1'), (3, 'userdef', 'XXX', 'comm')]
        frags_list = parse_fragment_file(mj_config_arr["fragments_filename"])
        self.fragments = []
        for i in frags_list:
            logging.info("Parsing fragment of type %s" % i[1])
            if i[1] == "matrix":
                self.fragments.append(mj_fragment_matrix(i[0]))
            elif i[1] == "lanthanide":
                self.fragments.append(mj_fragment_lanthanide(i[0],i[2]))
            elif i[1] == "equation":
                self.fragments.append(mj_fragment_equation(i[0],i[2]))
            elif i[1] == "userdef":
                self.fragments.append(mj_fragment_userdef(i[0]))
            else:
                raise Exception("mj_model.__init__ : grave internal error") # Should be impossible
    def parse(self, test=False):
        'Parse the model: just call parse() for all fragments. test=True omits actual calculation of matrices for matrix type fragments.'
        for i in self.fragments:
            i.parse(test=test)
    def save(self):
        'Stores the parsed model to disk'
        for i in self.fragments:
            i.save()
    def read(self):
        'Reads the stored model from disk'
        for i in self.fragments:
            i.read()
        srcf = open('source.dat')
        self.source = []
        self.temps = []
        for i in srcf.readlines():
            s = i.strip()
            if s == "" or s[:1] == "#":
                pass
            else:
                t,xt = s.split()
                self.source.append((float(t),float(xt)))
                self.temps.append(float(t))
    def get_source(self):
        return self.source
    def get_params_list(self):
        p_list = []
        for i in self.fragments:
            p_list.extend(i.get_params_list())
        return p_list
    def parse_fit_file(self, fin):
        '''Parsing previously stored fit and updating GUI accordingly. File opened by tkFileDialog is expected.'''
        finlines = fin.readlines()
        result = []
        values_dict = {}
        extvalues_dict = {}
        weights_dict = {}
        for i in finlines:
            j = i.strip()
            if j[0:3] == "#--":
                # "#--" start marks the line with a parameter
                try:
                    line = j[4:].strip() # Cutting the "#--" marker and the subsequent whitespace symbols
                    param_str, value_str = line.split("=")
                    param = param_str.strip()
                    value = float(value_str)
                    # First we will check if the parameter is a fragment weight -- like 'weight[0]'
                    match = re.match('weight\[(\d+)\]', param)
                    if match:
                        groups = match.groups() # A tuple with only one element containing the fragment number is expected, like ('0',)
                        if len(groups) != 1:
                            raise Exception("Error in parsing fragments weights -- probably malformed parameter: " + param)
                        frag_number = int(groups[0])
                        if frag_number in weights_dict.keys():
                            raise Exception("Weight of fragment %i declared more than once." % frag_number)
                        weights_dict[frag_number] = value
                    # Second, we will check if the parameter name matches a name of one of the external parameters
                    elif param in ['H', 'ddH', 'zJ', 'zJ_g', 'par_p', 'par_s', 'par_g', 'mass_corr', 'tip']:
                        if param in extvalues_dict.keys():
                            raise Exception("Parameter %s declared more than once." % param)
                        extvalues_dict[param] = value
                    # Maybe this is the stored R2 value?
                    elif param == "R2":
                        R2 = float(value)
                    # If neither above mentioned condition was met, it should be 'regular' parameter
                    else:
                        if param in values_dict.keys():
                           raise Exception("Parameter %s declared more than once." % param)
                        values_dict[param] = value
                except Exception as e:
                    raise Exception("Failure while parsing parameters-containing line of the stored fit file: "+str(e))
            elif j[0:1] == "#" or j == "":
                # Any other #-started lines are considered comments and ignored.
                # Empty strings are ignored as well
                pass
            else:
                try:
                    t_str, xt_str = j.split()
                    result.append((float(t_str),float(xt_str)))
                except Exception as e:
                    raise Exception("Failure while parsing a line in the stored fit file: "+str(e))
        return values_dict, extvalues_dict, weights_dict, result
    def get_xT(self, weights, ext_params, int_params):
        'Launches xT calculation over the entire model'
    # Just a stub now!

        H = ext_params['H']
        ddH = ext_params['ddH']
        zJ = ext_params['zJ']
        zJ_g = ext_params['zJ_g']
        par_p = ext_params['par_p']/100.0
        par_s = ext_params['par_s']
        par_g = ext_params['par_g']
        mass_corr = ext_params['mass_corr']
        tip = ext_params['tip']
        int_params_HdH = int_params+"H="+str(H)+os.linesep+"ddH="+str(ddH) #Extra os.linesep is redundant
        x_uncorr_arrays = [ i.chi(self.temps,int_params_HdH) for i in self.fragments ]
        xT_arr = []
        for t in range(len(self.temps)):
            x_values_at_t = [i[t][1] for i in x_uncorr_arrays] 
            x_uncorr = sum([x_values_at_t[i]*weights[i] for i in range(len(x_values_at_t))])
            xT_mf = self.temps[t]*x_uncorr/(1-zJ*x_uncorr*zJ_const/(zJ_g*zJ_g))/mass_corr
            xT = (1-par_p)*xT_mf+Nbk*par_g*par_g*par_p*par_s/(par_s+1)+tip*self.temps[t]
            xT_arr.append((self.temps[t],xT))
        print xT_arr
        return xT_arr
