#!/usr/bin/env pypy

###############################################################################
# Mjöllnir -- a simulator of the magnetic properties of polynuclear complexes
# Copyright (c) 2009--2020 Anton S. Lytvynenko, Elena A. Mikhalyova
#
# The authors thanks to Dr. V. A. Bogaenko and to Sergei Litvinenko for
# technical help and valuable discussion.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################

from mj_model_class import *

import os, re
from Tkinter import *
import tkFileDialog, tkMessageBox

class mj_params_record:
    def __init__(self,container,unit_name,unit_text,params_values):
        try:
            in_value = params_values[unit_name]
        except Exception:
            in_value = ""
        self.unit_name = unit_name
        self.unit_text = unit_text
        self.frame = Frame(container)
        self.w_l = Label(self.frame,text=self.unit_text)
        self.w_e = Entry(self.frame)
        self.w_e.insert(0,in_value)
        self.w_l.pack()
        self.w_e.pack()
        self.frame.pack()
    def get_value(self):
        return float(self.w_e.get())
    def set_value(self, params_values):
        try:
            in_value = params_values[self.unit_name]
            self.w_e.delete(0,END)
            self.w_e.insert(0,in_value)
        except KeyError as err:
            raise Exception("Parameter %s was found in the read file but absent in the model" % str(err))

class mj_params_frame:
    def __init__(self,mj_top_window,params_values={}):
        self.mj_top_window = mj_top_window
        self.int_params = self.mj_top_window.params_list
        self.container = Frame(self.mj_top_window.control_frame)
        self.params_widgets = []
        for i in self.int_params:
            self.params_widgets.append(mj_params_record(self.container,i,i,params_values))
        self.container.pack(side=TOP)
    def get_values(self):
        values_dict = {}
        for i in self.params_widgets:
            values_dict[i.unit_name]=i.get_value()
        return values_dict
    def set_values(self, values_dict):
        for i in self.params_widgets:
            i.set_value(values_dict)

class mj_weights_frame:
    def __init__(self,mj_top_window,weights={}):
        self.mj_top_window = mj_top_window
        self.model = mj_top_window.model
        self.container = Frame(self.mj_top_window.control_frame)
        self.params_widgets = []
        if len(weights)==0:
            for i in range(len(self.model.fragments)):
                self.params_widgets.append(mj_params_record(self.container, i, "Weight of fragment %i" % i,{i:'1.0'}))
                print i
        else:
            for i in range(len(weights)):
                self.params_widgets.append(mj_params_record(self.container, i, "Weight of fragment %i" % i, {i:weights[i]}))
        self.container.pack(side=TOP)
    def get_values(self):
        values_dict = {}
        for i in self.params_widgets:
            values_dict[i.unit_name]=i.get_value()
        return values_dict
    def set_values(self, values_dict):
        for i in self.params_widgets:
            i.set_value(values_dict)

class mj_extparams_frame:
    def __init__(self,mj_top_window,ext_params_values={}):
        self.mj_top_window = mj_top_window
        self.ext_params = {"H":"H","ddH":"-lg(dH/H)","zJ":"zJ\'","zJ_g":"Average g for molecular field",
                            "par_p":'molar % of paramagnetic impurity',"par_s":"paramagnetic impurity spin",
                            "par_g":"g of the paramagnetic impurity","mass_corr":"Molar mass correction","tip":"TIP"}
        if len(ext_params_values.keys())==0:
            self.ext_params_values = {"H":2.0,"ddH":4.0,"zJ":0.0,"zJ_g":2.0,"par_p":0.0,"par_s":0.5,"par_g":2.0,
                                        "mass_corr":1.0,"tip":0.0}
        self.container = Frame(self.mj_top_window.control_frame)
        self.params_widgets = []
        for i in self.ext_params.keys():
            self.params_widgets.append(mj_params_record(self.container,i,self.ext_params[i],self.ext_params_values))
        self.container.pack(side=TOP)
    def get_values(self):
        values_dict = {}
        for i in self.params_widgets:
            values_dict[i.unit_name]=i.get_value()
        return values_dict
    def set_values(self, values_dict):
        for i in self.params_widgets:
            i.set_value(values_dict)

def R2(experim,calcd):
    numerator = 0
    denominator = 0
    for i in range(len(calcd)):
        numerator+=(calcd[i][1]-experim[i][1])*(calcd[i][1]-experim[i][1])
        denominator+=(experim[i][1])*(experim[i][1])
    return numerator/denominator

class mj_canvas_frame:
    def __init__(self,mj_top_window):
        self.model = mj_top_window.model
        self.mj_top_window = mj_top_window
        self.root = mj_top_window.root
        self.container = Frame(self.root)

        # Creating Canvas object for plots
        self.plot_width=int(0.75*self.root.winfo_screenwidth())
        self.plot_height=int(0.75*self.root.winfo_screenheight())

        self.R2 = Label(self.container,text="R2 = ")
        self.R2.pack(side=TOP)
        self.plot = Canvas(self.container, background="white", width=self.plot_width, height=self.plot_height)
        self.plot.pack(side=TOP)
        self.cnv_ctrl_frame = Frame(self.container)
        Label(self.cnv_ctrl_frame, text="Min T:").pack(side=LEFT)
        self.minT_entry = Entry(self.cnv_ctrl_frame)
        self.minT_entry.insert(0,'0')
        self.minT_entry.pack(side=LEFT)
        Label(self.cnv_ctrl_frame, text="Max T:").pack(side=LEFT)
        self.maxT_entry = Entry(self.cnv_ctrl_frame)
        self.maxT_entry.insert(0,'300')
        self.maxT_entry.pack(side=LEFT)
        Label(self.cnv_ctrl_frame, text="Min xT:").pack(side=LEFT)
        self.minxT_entry = Entry(self.cnv_ctrl_frame)
        self.minxT_entry.insert(0,'0')
        self.minxT_entry.pack(side=LEFT)
        Label(self.cnv_ctrl_frame, text="Max xT:").pack(side=LEFT)
        self.maxxT_entry = Entry(self.cnv_ctrl_frame)
        self.maxxT_entry.insert(0,'5')
        self.maxxT_entry.pack(side=LEFT)

        self.set_bounds()

        Button(self.cnv_ctrl_frame,text="Set",command=self.update_bounds).pack(side=LEFT)
        self.cnv_ctrl_frame.pack(side=BOTTOM,fill=X)
        self.container.pack(side=LEFT,expand=YES,fill=BOTH)

        self.small_line_size = 10
        self.plot_padding = 75

        self.source = self.model.get_source()

        self.draw_plot()

    def phys_x(self,T):
        return self.plot_padding+((T-self.minT)/(self.maxT-self.minT))*(self.plot_width-2*self.plot_padding)
    def phys_y(self,xT):
        h = self.plot_height
        return h - (self.plot_padding+((xT-self.minxT)/(self.maxxT-self.minxT))*(h-2*self.plot_padding))
    def draw_plot(self):
#        canvas_width = self.plot.winfo_width()
#        canvas_height = self.plot.winfo_height()
        '''Refreshes the plot'''
        self.plot.delete("axes") # We delete all old axes lines in order to redraw them
        self.plot.delete("source")
        self.plot.create_line(self.phys_x(self.minT),self.phys_y(self.minxT),
                self.phys_x(self.maxT),self.phys_y(self.minxT),tags="axes")  # 0x line
        self.plot.create_line(self.phys_x(self.minT),self.phys_y(self.minxT),
                self.phys_x(self.minT),self.phys_y(self.maxxT),tags="axes")  # 0y line

        # Now we are drawing ticks and tick labels on 0x axis:
        t_curr = self.minT
        while t_curr<self.maxT:
            self.plot.create_line(self.phys_x(t_curr),self.phys_y(self.minxT)-self.small_line_size,
                                    self.phys_x(t_curr),self.phys_y(self.minxT)+self.small_line_size,tags="axes")
            self.plot.create_text(self.phys_x(t_curr),self.phys_y(self.minxT)+2*self.small_line_size,
                                    text=format(t_curr,".1f"),anchor=N,tags="axes")
            t_curr+=self.intT

        # And 0y axis:
        xt_curr = self.minxT
        while xt_curr<self.maxxT:
            self.plot.create_line(self.phys_x(self.minT)-self.small_line_size,self.phys_y(xt_curr),
                                    self.phys_x(self.minT)+self.small_line_size,self.phys_y(xt_curr),tags="axes")
            self.plot.create_text(self.phys_x(self.minT)-2*self.small_line_size,self.phys_y(xt_curr),
                                    text=format(xt_curr,".2f"),anchor=E,tags="axes")
            xt_curr+=self.intxT
        # Drawing source line:
        source = self.source[0:]
        x,xt = source.pop(0)
        while len(source)>0:
            x_next,xt_next = source.pop(0)
            self.plot.create_line(self.phys_x(x),self.phys_y(xt),self.phys_x(x_next),self.phys_y(xt_next),
                                    tags="source",fill="red")
            x = x_next
            xt = xt_next

    def draw_result(self,result):
        res = result[0:]
        x,xt = res.pop(0)
        while len(res)>0:
            x_next,xt_next = res.pop(0)
            self.plot.create_line(self.phys_x(x),self.phys_y(xt),self.phys_x(x_next),self.phys_y(xt_next),
                                    tags="result",fill="green")
            x = x_next
            xt = xt_next
        self.R2.config(text="R2 = "+format(R2(self.source,result),".3g"))

    def set_bounds(self):
        '''This method is called when Rescale button pressed in order to check and set new scale of the plot'''
        minT = float(self.minT_entry.get())
        maxT = float(self.maxT_entry.get())
        minxT = float(self.minxT_entry.get())
        maxxT = float(self.maxxT_entry.get())
        if minT>=maxT or minxT>=maxxT:
            raise Exception('mj_canvas_frame.draw_plot:invalid values of canvas limits')
        self.minT=minT
        self.maxT=maxT
        self.minxT=minxT
        self.maxxT=maxxT
        self.intT = (self.maxT-self.minT)/6.0
        self.intxT = (self.maxxT-self.minxT)/6.0
    def update_bounds(self):
        self.set_bounds()
        self.draw_plot()
class mj_window:
    def __init__(self):
        self.model = mj_model()
        self.model.read()
        self.params_list = self.model.get_params_list()
        self.source = self.model.get_source()

        self.root = Tk()
        screen_width = self.root.winfo_screenwidth()
        screen_height = self.root.winfo_screenheight()

        self.canvas_frame = mj_canvas_frame(self)

        self.control_frame = Frame(self.root)
        self.params_frame = mj_params_frame(self)
        self.extparams_frame = mj_extparams_frame(self)
        self.weights_frame = mj_weights_frame(self)
        # Main Button
        self.start_button = Button(self.control_frame, text="Calculate", command=self.calc_xt)

        self.start_button.pack(side=BOTTOM)

        self.save_button = Button(self.control_frame, text="Save the last fit", state=DISABLED, command=self.save_result)
        self.save_button.pack(side=BOTTOM)

        self.read_button = Button(self.control_frame, text="Read stored fit", command=self.parse_fit_file_gui)
        self.read_button.pack(side=BOTTOM)

        self.control_frame.pack(side=RIGHT)
        self.root.mainloop()

    def extract_params(self):
        '''Aux function for extracting parameters values from the GUI and storing it into the internal object variables.'''
        self.values_dict = self.params_frame.get_values()
        self.extvalues_dict = self.extparams_frame.get_values()
        self.weights_dict = self.weights_frame.get_values()

    def save_result(self):
        '''Select file and write the current simulation curve together with its simulation parameters into the file.'''
        fout = tkFileDialog.asksaveasfile()
        if fout:
            fout.write("#-- R2 = " + str(self.R2) + os.linesep)
            fout.write("### Parameters ###"+os.linesep)
            fout.write(os.linesep.join([ "#-- %s = %f" % (i, self.values_dict[i]) for i in self.values_dict.keys()]))
            fout.write(os.linesep+"### Extended parameters ###"+os.linesep)
            fout.write(os.linesep.join([ "#-- %s = %f" % (i, self.extvalues_dict[i]) for i in self.extvalues_dict.keys()]))
            fout.write(os.linesep+"### Fragments weights ###"+os.linesep)
            fout.write(os.linesep.join([ "#-- weight[%i] = %f" % (int(i), self.weights_dict[i]) for i in self.weights_dict.keys()]))
            fout.write(os.linesep+"### T    xT"+os.linesep)
            fout.write(os.linesep.join([ "%f    %f"%(t[0], t[1]) for t in self.result]))
            fout.close()

    def parse_fit_file_gui(self):
        try:
            fin = tkFileDialog.askopenfile()
            if fin:
                values_dict, extvalues_dict, weights_dict, result = self.model.parse_fit_file(fin)
                self.params_frame.set_values(values_dict)
                self.extparams_frame.set_values(extvalues_dict)
                self.weights_frame.set_values(weights_dict)
                # If the above instructions succeeded, let's set the value of R2
                self.R2 = R2
                if result:
                    self.result = result
                    self.canvas_frame.draw_result(self.result)
        except Exception as e:
            tkMessageBox.showerror("Failure while parsing a fit file", "Exception was raised during the fit file parsing, details follow:" + os.linesep + os.linesep + str(e))

    def calc_xt(self):
        '''Prepare everything, call the solver of the associated model object, store results and redraw canvas.'''
        try:
            self.extract_params()
            params_script = ''
            for i in self.values_dict.keys():
                params_script+=i+"="+str(self.values_dict[i])+os.linesep
            self.result = self.model.get_xT(self.weights_dict,self.extvalues_dict,params_script)
            self.R2 = R2(self.source, self.result)
            self.canvas_frame.draw_result(self.result)
            self.save_button.config(state=NORMAL) # Saving results should be enabled only after at least one simulation was performed.
        except Exception as e:
            self.save_button.config(state=DISABLED)
            tkMessageBox.showerror("Simulation failed", "Exception was raised during the simulation, details follow:"+os.linesep+os.linesep+str(e))

m = mj_window()
