#!/usr/bin/env pypy
# -*- coding: utf-8 -*-

###############################################################################
# Mjöllnir -- a simulator of the magnetic properties of polynuclear complexes
# Copyright (c) 2009--2020 Anton S. Lytvynenko, Elena A. Mikhalyova
#
# The authors thanks to Dr. V. A. Bogaenko and to Sergei Litvinenko for
# technical help and valuable discussion.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################

import os, subprocess, threading

from Tkinter import *
import tkFileDialog

class mjollnir_launcher():
    '''A class incapsulating the entire mjollnir_launcher tkinter window and its handlers.'''
    def __init__(self):
        '''Creates UI, initializes reasonable defaults.'''
        self.working_dir = os.getcwd()
        self.mj_dir = os.path.dirname(os.path.realpath(__file__))
        self.root=Tk()
        self.root.title("Mjöllnir launcher")

        self.working_dir_label = Label(self.root, text="Current working directory: %s" % self.working_dir)
        self.working_dir_label.grid(column=0, row=0, sticky=W+E+N+S, columnspan=3)

        self.cwd_button = Button(self.root, text="Select working directory", command=self.selectfolder)
        self.cwd_button.grid(column=0, row=1, sticky=W+E+N+S)
        self.launch_modelparse_button = Button(self.root, text="Launch model parsing", command=self.modelparse_launcher)
        self.launch_modelparse_button.grid(column=1, row=1, sticky=W+E+N+S)
        self.launch_fitint_button = Button(self.root, text="Launch fitting", command=self.fitint_launcher)
        self.launch_fitint_button.grid(column=2, row=1, sticky=W+E+N+S)

        self.subprocess_running = False

        self.listing = ""
        self.listing_widget = Text(self.root)
        self.listing_widget.grid(column=0, row=2, columnspan=3)
        self.listing_widget.after(200,self.update_controls)
        self.root.mainloop()
        
    def selectfolder(self):
        '''Method for current working directory selection.'''
        w = tkFileDialog.askdirectory()
        if (w):
            self.working_dir = w
            self.working_dir_label.config(text="Current working directory: %s" % self.working_dir)
            os.chdir(self.working_dir)

    def modelparse(self):
        '''Launches mjollnir_parse.py script in a child process, tracks the output (appending it to self.listing field) and reads the return value.
           Intended to be launched in a separate thread to avoid UI freezes, as model parsing is time-consuming for large models.
           Sets self.subprocess_running flag to block UI controls until the child process yields the result in order to avoid mess.'''
        p = subprocess.Popen(os.path.join(self.mj_dir, "mjollnir_parse.py"), stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        self.subprocess_running = True
        self.listing+="[Model parsing started]"+os.linesep
        # Suggested by https://www.endpoint.com/blog/2015/01/28/getting-realtime-output-using-python
        while True:
            l = p.stdout.readline()
            if l == "" and p.poll() is not None:
                break
            if l!="":
                self.listing+=l
        if p.poll() != 0:
            self.listing+="[Model parsing failed]"+os.linesep
        else:
            self.listing+="[Model parsing succeeded]"+os.linesep
        self.subprocess_running = False

    def modelparse_launcher(self):
        '''A few extra lines for launching modelparse in a separate thread.'''
        t = threading.Thread(target=self.modelparse)
        t.start()

    def fitint(self):
        self.subprocess_running = True
        self.listing+="[Starting fitint]"+os.linesep
        try:
            p = subprocess.check_output(os.path.join(self.mj_dir,"mjollnir_fitint.py"))
            self.listing+="[Fitint exited gracefully]"+os.linesep
        except subprocess.CalledProcessError as e:
            self.listing+="[Fitint failed with code %i]"%e.returncode+os.linesep
            self.listing+=e.output+os.linesep
        finally:
            self.subprocess_running = False
        
    def fitint_launcher(self):
        '''A few extra lines for launching fitint in a separate thread.'''
        t = threading.Thread(target=self.fitint)
        t.start()
    
    def update_controls(self):
        '''UI updater. Disables and enables buttons depending on self.subprocess_running flag, updates listing_widget (process log)
           and re-schedules its own repetition in future.'''
        if self.subprocess_running:
            self.launch_modelparse_button.config(state=DISABLED)
            self.launch_fitint_button.config(state=DISABLED)
            self.cwd_button.config(state=DISABLED)
        else:
            self.launch_modelparse_button.config(state=NORMAL)
            self.launch_fitint_button.config(state=NORMAL)
            self.cwd_button.config(state=NORMAL)
        self.listing_widget.config(state=NORMAL)
        self.listing_widget.delete(1.0, END)
        self.listing_widget.insert(END, self.listing)
        self.listing_widget.config(state=DISABLED)
        self.listing_widget.after(200,self.update_controls)

if __name__ == "__main__":
    launcher = mjollnir_launcher()
